#define __USE_GNU
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "astrotime.h"

AstroDateTime::AstroDateTime( void ){
    gettimeofday( &hightime, NULL );
    hightime.tv_sec -= 12*60*60;
    curr = localtime(&hightime.tv_sec);
    start_loc_year = 1900 + curr->tm_year;
    start_loc_month = curr->tm_mon + 1;
    start_loc_day = curr->tm_mday;          //Calendar date 12 hours ago as starting date

    gettimeofday( &hightime, NULL );
    curr = gmtime(&hightime.tv_sec);
    utc_year = 1900 + curr->tm_year;
    utc_month = curr->tm_mon + 1;
    utc_day = curr->tm_mday;
    curr->tm_hour = 0;
    curr->tm_min = 0;
    curr->tm_sec = 0;
    start_midnight = timegm( curr );
    _julian = julian( utc_year, utc_month, utc_day );
    _s0 = s0();
    srand( hightime.tv_usec );
};

void AstroDateTime::current(void) {
    gettimeofday( &hightime, NULL );
    if((long)hightime.tv_usec >= 999500L ) {
        hightime.tv_sec++;
        hightime.tv_usec = 0;
    }
    curr = gmtime(&hightime.tv_sec);
    int lday = (hightime.tv_sec - start_midnight)/(24*60*60);
    utc_hour = curr->tm_hour+24*lday;
    utc_min = curr->tm_min;
    utc_sec = curr->tm_sec;
    utc_msec = (int)(hightime.tv_usec+500)/1000 ;
}

int	AstroDateTime::year( void ) {
    return(utc_year);
}

int	AstroDateTime::month( void ) {
    return(utc_month);
}

int	AstroDateTime::day( void ) {
    return(utc_day);
}

int	AstroDateTime::hour( void ) {
    return(utc_hour);
}

int	AstroDateTime::min( void ) {
    return(utc_min);
}

int	AstroDateTime::sec( void ) {
    return(utc_sec);
}

int	AstroDateTime::msec( void ) {
    return(utc_msec);
}

int	AstroDateTime::julian( void ) {
    return(_julian);
}

int AstroDateTime::julian( int year, int month, int day ) {
    int result = (3057*month)/100 + (365*year) + 1721028L;
    result = result + (year/4 - year/100 + year/400 + day);
    if( month > 2 ) return( result);
    int j = 2;
    if( year%4 == 0 )   j -= 1;
    if( year%100 == 0 ) j -= 1;
    if( year%400 == 0 ) j += 1;
    result += j;
    return( result );
}

char *AstroDateTime::strTime( void ) {
    static char utcstring[16];
    sprintf( utcstring, "%02d:%02d:%02d.%03d", utc_hour, utc_min, utc_sec, utc_msec );
    return (utcstring);
}

char *AstroDateTime::strDate( void ) {
    static char utcstring[16];
    sprintf( utcstring, "%04d-%02d-%02d", start_loc_year, start_loc_month, start_loc_day );
    return (utcstring);
}

char *AstroDateTime::strFileName( void ) {
    static char utcstring[16];
    sprintf( utcstring, "%02d%02d%02d", start_loc_year-2000, start_loc_month, start_loc_day );
    return (utcstring);
}

/* Calculation of siderial time in the Greewhich midnigth (in days) */

double AstroDateTime::s0( void ) {
    double c3 = 1.0775925925925926E-06;
    double c2 = 2.7379093507870371E-03;
    double c1 = 0.2790572732638889;

    double d = (double)_julian - (double)2451545.5;
    double t2 = d/(double)36525.0;
    double result = c1 + c2*d + c3*t2*t2;
    return ( result - floor( result ) );
}

/* Calculation of the local siderial time  (in days) */

double AstroDateTime::currentSec( void ) {
    gettimeofday( &hightime, NULL );
    double lt = (double)(hightime.tv_sec - start_midnight) + hightime.tv_usec/1000000;
    return( lt );
}

double AstroDateTime::sidTime( double longitude ) {
    double  stg = _s0 + 1.0027379093*currentSec()/86400.0 + longitude;
    stg = stg - floor(stg);
    return( stg );
}
