
#define  _EXCHANGE
#include <sys/ioctl.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <time.h>
#include </usr/src/rs485/include/mcua.h>
#include </usr/src/rs485/include/rserrno.h>
#include <sys/time.h>
#include <pthread.h>

#include "astrotime.h"
#include "exchange.h"

namespace Signal {
    char *abbrev( int sc );
    char *message( int sc );
    int  type( int sc );

    Sig data[] = {
        { UND, LOOP_SIG, "UND", "Unexpected data in place of Signal "},
        { ACK, INFO_SIG, "ACK", "Successful data transmission"},
        { NAK, LOOP_SIG, "NAK", "Broken packet was received"},
        { NOD, INFO_SIG, "NOD", "Required data are not ready"},
        { ACN, INFO_SIG, "ACN", "This command does not exist"},
        { ACY, INFO_SIG, "ACY", "Acknowledgement of existing command"},
        { ACW, INFO_SIG, "ACW", "Command can't be executed right now"},
        { DSC, LOOP_SIG,  "DSC", "Module doesn't respond or timeout"},
        { COL, LOOP_SIG,  "COL", "Transmission error or packets collision"},
        { PBO, CHECK_SIG, "PBO", "LPT converter FIFO overflow"},
        { PKL, CHECK_SIG, "PKL", "Packet loss detected"},
        { EER, LOOP_SIG,  "EER", "General exchange error"},
        { 0  , ABORT_SIG, "???", "Unknown signal" } };

    int num = sizeof(data)/sizeof(data[0]) - 1;
    int index;
}

int Signal::type( int sc ) {
    int i;
    if( sc <= MAX_PACKET_SIZE ) return( INFO_SIG );
    for( i = 0; i < num; i++ ) if( data[i].code == sc ) break;
    index = i;
    return( data[i].stype );
}

char* Signal::abbrev( int sc = 0 ) {
    int i = index;
    if( sc != 0 ) for( i = 0; i < num; i++ ) if( data[i].code == sc ) break;
    return( data[i].abbr );
}

char* Signal::message( int sc = 0 ) {
    int i = index;
    if( sc != 0 ) for( i = 0; i < num; i++ ) if( data[i].code == sc ) break;
    return( data[i].text );
}

extern AstroDateTime utc;

namespace {

    pthread_mutex_t m485 = PTHREAD_MUTEX_INITIALIZER;


    int     RS485_transaction( unsigned char*, int, unsigned char* );
    int     checkErrors( int, int );
    void    terminateExchange( void );
    const char* RS_strerror( int );
    int     micro_delay( int );
    void    logFileOpen( void );

    FILE    *logfile = 0;

    unsigned char strcomm[IO_BUFF_SIZE];
    unsigned char strread[IO_BUFF_SIZE];
    unsigned char strdata[IO_BUFF_SIZE];

    int extranum = 0;
    int ret_signal;
    int id_mcua = -1;

    const char* RS_strerror(int errnum) {
        switch(errnum) {
            case ERSNODEVICE    :
            case ENOENT         : return "no device mcua0";
            case ERSNOPORT      : return "no this port";
            case ERSNOINTERFACE : return "no RS485 interface";
            case ENODEV         :
            case ERSNODRIVER    : return "no RS485 driver";
            case ERSBUSY        : return "RS485 device is busy";
            case ERSSHUTDOWN    : return "device is off or disconnected";
            case ERSISATMOUT    : return "RS485 ISA time out";
            case ERSNOCONVREPLY : return "no reply from RS485 converter";
            case ERSFRAME       : return "RS485 frame error";
            case ERSFIFOVERFLOW : return "RS485 FIFO overflow";
            case ERSNOIOCTL     : return "RS485: no such ioctl command";
            default             : if( errnum < ERSBASE ) return strerror(errnum);
                                  return "unknown RS485 error";
        }
    }

    int micro_delay( int mks ) {
        struct timeval moms;
        struct timeval mome;
        int mksd = 0;
        int secd = 0;
        gettimeofday( &moms, NULL );
        while( mksd < mks ) {
            gettimeofday( &mome, NULL );
            secd = mome.tv_sec - moms.tv_sec;
            mksd = mome.tv_usec - moms.tv_usec;
            if( secd < 0 || secd > 1 ) return( -1);
            if( secd > 0 ) mksd += 1000*1000;
        }
        return( 0 );
    }

    int micro_wait( int mks ) {
        static struct timeval moms;
        struct timeval mome;
        int mksd = 0;
        int secd = 0;
        if( mks == 0 ){
            gettimeofday( &moms, NULL );
            return( 0 );
        }
        while( mksd < mks ) {
            gettimeofday( &mome, NULL );
            secd = mome.tv_sec - moms.tv_sec;
            mksd = mome.tv_usec - moms.tv_usec;
            if( secd < 0 || secd > 1 ) return( -1);
            if( secd > 0 ) mksd += 1000*1000;
        }
        return( 0 );
    }

    int checkErrors( int nsent, int noerr ) {
        char string[128];
        if( noerr || logfile == 0 ) return( 0 );        // Only error message
        utc.current();
        strcpy( string, utc.strTime() );
        int point = strlen( string );
        strcat(string+point," > ");
        point += 3;
        for( int i = 0; i < nsent; i++ ) point += sprintf( string+point, "%.2X ", strcomm[i] );
        strcat(string+point," <-- ");
        point += 5;
        if( ret_signal <= MAX_PACKET_SIZE )
            for( int i = 0; i < ret_signal; i++ ) point += sprintf( string+point, "%.2X ", strread[i] );
        else  strcat( string+point, Signal::abbrev() );
        fprintf( logfile, "%s\n", string );
        return(0);
    }

    const char *auth = "RS485 exchange";

    int RS485_transaction( unsigned char *buffi, int nbyte, unsigned char *buffo ){
        int att = MAXRETRY;
        char temporal[MAX_PACKET_SIZE];
        int reply, rules;
        while( att-- ) {
            memcpy( temporal, buffi, nbyte );
            micro_wait( 1500 );
            reply = write( id_mcua, (char*)temporal, nbyte );
            micro_wait(0);
            if( reply < 0 ) {
                ioctl( id_mcua, RS_RESET ); //Check!!!!
                micro_delay( 4000 );
                if( !att ) throw Error::Exchange( auth, RS_strerror( errno ) );
                else fprintf( stderr, "%s: %s (errno %d)\n", auth, RS_strerror( errno ), errno );
                continue;
            }
            rules = Signal::type( reply );
            if( rules == INFO_SIG ) break;
            if( rules == LOOP_SIG ) continue;
            if( rules == CHECK_SIG ) {
                fprintf( stderr, "%s\n", Signal::message());
                break;
            }
            if( rules == ABORT_SIG ) throw Error::Exchange( auth, Signal::message() );
        }
        if( reply <= MAX_PACKET_SIZE ) memcpy( buffo, temporal+1, --reply );
        return( reply );
    }

    void terminateExchange( void ) {
        if( id_mcua  > 0) close( id_mcua );
    }

    void logFileOpen( void ){
        char name[16];
        strcpy( name, utc.strFileName() );
        strcat( name, ".elog" );
        logfile = fopen( name, "a");
	utc.current();
        if(logfile ) fprintf( logfile, "Session %s %s\n", utc.strDate(), utc.strTime() );
        else perror("\nRS485 exchange protocol file: ");
    }

}

void RS485::initExchange( void ) {
    const char *from = "RS485 initialization";
    char vers[18];
            if( id_mcua  > 0) close( id_mcua );
    atexit( terminateExchange );
    if( (id_mcua = open( DEV_FILE_OUT, O_RDWR )) < 0 ) { // 
        throw Error::Exchange( from, RS_strerror( errno ) );
    }
    int verscode = ioctl( id_mcua, RS_GET_VERS, vers );
    if( verscode < 0 ) {
        throw Error::Exchange( from, "unsupported driver version" );
    }
    if(RS485_RTL_CODE(verscode)) strcat( vers, "-rtl" );
    fprintf( stdout, "RS485 initialization with driver ver.%s done\n", vers );
    logFileOpen();
}

int RS485::askByte( int module, int command, char *avalue ){
    int nsent = 0, good = 0;
    pthread_mutex_lock(&m485);
    strcomm[nsent++] = (char)module;
    strcomm[nsent++] = (char)command;                 //Context analys!!!
    ret_signal = RS485_transaction( strcomm, nsent, strread );
    if( ret_signal == 1 ) {
        *avalue = (int)(*strread);
        good = 1;
    }
    checkErrors( nsent, good );
    pthread_mutex_unlock(&m485);
    return( good );
}

int RS485::askWord( int module, int command, short *avalue ){
    int nsent = 0, good = 0;
    pthread_mutex_lock(&m485);
    strcomm[nsent++] = (char)module;
    strcomm[nsent++] = (char)command;                 //Context analys!!!
    ret_signal = RS485_transaction( strcomm, nsent, strread );
    if( ret_signal == 2 ) {
        memcpy( (char*)avalue, strread, 2 );
        good = 1;
    }
    checkErrors( nsent, good );
    pthread_mutex_unlock(&m485);
    return( good );
}

int RS485::askLong( int module, int command, int *avalue ){
    int nsent = 0, good = 0;
    pthread_mutex_lock(&m485);
    strcomm[nsent++] = (char)module;
    strcomm[nsent++] = (char)command;                 //Context analys!!!
    ret_signal = RS485_transaction( strcomm, nsent, strread );
    if( ret_signal == 3 ) {
        strread[3] = ((char)strread[2] < 0 ? 0xff : 0 );
        memcpy( (char*)avalue, strread, 4 );
        good = 1;
    }
    checkErrors( nsent, good );
    pthread_mutex_unlock(&m485);
    return( good );
}

int RS485::askData( int module, int command, unsigned char *data){
    int nsent = 0, good = 0;
    pthread_mutex_lock(&m485);
    strcomm[nsent++] = (char)module;
    strcomm[nsent++] = (char)command;                 //Context analys!!!
    ret_signal = RS485_transaction( strcomm, nsent, strread );
    if( ret_signal <= MAX_PACKET_SIZE ) {
        memcpy( data, strread, ret_signal );
        good = ret_signal;
    }
    if( ret_signal == NOD ) good = -1;
    checkErrors( nsent, good );
    pthread_mutex_unlock(&m485);
    return( good );
}

int RS485::sendSimpleCommand( int module, int command ){
    int nsent = 0, good = 0;
    pthread_mutex_lock(&m485);
    strcomm[nsent++] = (char)module;
    strcomm[nsent++] = (char)command;                 //Context analys!!!
    ret_signal = RS485_transaction( strcomm, nsent, strread );
    if( ret_signal == ACY || ret_signal == ACW  ) good = 1;
    checkErrors( nsent, good );
    pthread_mutex_unlock(&m485);
    return( good );
}

int RS485::sendByteCommand( int module, int command, char arg){
    int nsent = 0, good = 0;
    pthread_mutex_lock(&m485);
    strcomm[nsent++] = (char)module;
    strcomm[nsent++] = (char)command;                 //Context analys!!!
    strcomm[nsent++] = arg;
    ret_signal = RS485_transaction( strcomm, nsent, strread );
    if( ret_signal == ACY || ret_signal == ACW  ) good = 1;
    checkErrors( nsent, good );
    pthread_mutex_unlock(&m485);
    return( good );
}

int RS485::sendWordCommand( int module, int command, short arg){
    int nsent = 0, good = 0;
    pthread_mutex_lock(&m485);
    strcomm[nsent++] = (char)module;
    strcomm[nsent++] = (char)command;                 //Context analys!!!
    strcomm[nsent++] = low(arg);
    strcomm[nsent++] = high(arg);
    ret_signal = RS485_transaction( strcomm, nsent, strread );
    if( ret_signal == ACY || ret_signal == ACW  ) good = 1;
    checkErrors( nsent, good );
    pthread_mutex_unlock(&m485);
    return( good );
}

int RS485::sendData( int module, int length, unsigned char *data){
    int nsent = 0, good = 0;
    pthread_mutex_lock(&m485);
    strcomm[nsent++] = (char)module;
    strcomm[nsent++] = (char)length;                 //Context analys!!!
    memcpy(strcomm+2, data, length );
    nsent = length+2;
    ret_signal = RS485_transaction( strcomm, nsent, strread );
    if( ret_signal == ACY || ret_signal == ACW || ret_signal == ACK ) good = 1;
    checkErrors( nsent, good );
    pthread_mutex_unlock(&m485);
    return( good );
}

int RS485::sendVoid( unsigned char *buff, int length, unsigned char *data ){
    int nsent = length, good = 0;
    pthread_mutex_lock(&m485);
    memcpy( strcomm, buff, length );
    ret_signal = RS485_transaction( strcomm, nsent, strread );
    if( ret_signal <= MAX_PACKET_SIZE ) {
        memcpy( data, strread, ret_signal );
        good = ret_signal;
    } else strcpy( (char*)data, Signal::abbrev(ret_signal) );
    checkErrors( nsent, good );
    pthread_mutex_unlock(&m485);
    return( good );
}

int RS485::deviceBusy( void ){
    if( ret_signal == ACY ) return(0);
    if( ret_signal == ACW ) return(1);
    return(-1);
}

int RS485::readData( unsigned char *head, unsigned char *data){
    int n_read_data = 0;
    int nsent = 0, good = 0;
    pthread_mutex_lock(&m485);
    if( n_read_data = read( id_mcua, strdata, IO_BUFF_SIZE ) ) {
        good = n_read_data-4;
        memcpy( head, strdata+1, 4 );
        memcpy( data, strdata+5, n_read_data-4 );
    }
    pthread_mutex_unlock(&m485);
    return( good );
}

int RS485::readData( int *from, int *numb, unsigned short *data){
    int n_read_data = 0;
    unsigned short pnum;
    int nsent = 0, good = 0;
    pthread_mutex_lock(&m485);
    if( n_read_data = read( id_mcua, strdata, IO_BUFF_SIZE ) ) {
        good = (n_read_data-4)/2;
        *from = (int)*(strdata+2);
        memcpy( &pnum, strdata+3, 2 );
        *numb = pnum + extranum;
        if( pnum == 0xFFFF ) extranum += 0x10000;
        memcpy( data, strdata+5, n_read_data-4 );
    }
    pthread_mutex_unlock(&m485);
    return( good );
}

void RS485::resetData( void ){
    pthread_mutex_lock(&m485);
    extranum = 0;
    ioctl( id_mcua, RS_BLNUMBER_RESET );
    pthread_mutex_unlock(&m485);
}

void RS485::setBaudeRate( int speed ){
    pthread_mutex_lock(&m485);
    ioctl( id_mcua, RS_SET_BAUDRATE, speed );
    pthread_mutex_unlock(&m485);
}
